/*

 usb.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "usb.h"
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>

#include "crc8.h"
#include "ideservd.h"
#include "log.h"

#ifndef WIN32
#include <ftdi.h>

static struct ftdi_context *ftDevice;
#else
#include <windows.h>	//also in ideservd.h, but we need it now
#include "ftd2xx.h"

FT_STATUS ftStatus;
FT_HANDLE ftHandle;
#endif //WIN32

static const char *i_dev;
static int inited;
static unsigned char ebufo[4096];
static unsigned int ebufop;

static int flush(void);

static volatile int tout;
static void timeout(int UNUSED(x)) { tout = 1; }

static const char *drivername(void) {
    return "USB";
}

static int initialize(const char *dev, int UNUSED(unused), Log *lf, int lastfail) {
#ifndef WIN32
    struct ftdi_device_list *list, *i, *j;
    char desc[256], serial[256];

    i_dev = dev;
    inited = 0;

    ftDevice = ftdi_new();
    if (!ftDevice) {
        if (lastfail != -1) log_print(lf, "Out of memory\n");
        return -1;
    }

    if (ftdi_usb_find_all(ftDevice, &list, 0x0403, 0x6001) < 0) {
        if (lastfail != -2) log_print(lf, "ftdi_usb_find_all: %s\n", ftdi_get_error_string(ftDevice));
        ftdi_free(ftDevice);
        ftDevice = NULL;
        return -2;
    }

    for (i = list; i; i = i->next) {
        if (ftdi_usb_get_strings(ftDevice, i->dev, NULL, 0, desc, sizeof(desc), serial, sizeof(serial)) < 0) {
            continue;
        }
        if (dev) {
            if (!strcmp(dev, serial)) break;
        } else {
            if (!strcmp(desc, "IDE64 USB DEVICE")) break;
        }
    }
    if (!i && !dev) {
        for (i = list; i; i = i->next) {
            if (ftdi_usb_get_strings(ftDevice, i->dev, NULL, 0, desc, sizeof(desc), serial, sizeof(serial)) < 0) {
                continue;
            }
            if (!strcmp(desc, "FT245R USB FIFO")) break;
        }
    }
    for (j = list; j; j = j->next) {
        char desc2[256], serial2[256];
        if (i == j) {
            continue;
        }
        if (ftdi_usb_get_strings(ftDevice, j->dev, NULL, 0, desc2, sizeof(desc2), serial2, sizeof(serial2)) < 0) {
            continue;
        }
        log_print(lf, "Ignored device: %s (%s)\n", serial2, desc2);
    }
    if (!i) {
        ftdi_list_free(&list);
        if (lastfail != -3) log_print(lf, "No usable USB device found for now\n");
        ftdi_free(ftDevice);
        ftDevice = NULL;
        return -3;
    }
    if (ftdi_usb_open_dev(ftDevice, i->dev)) {
        ftdi_list_free(&list);
        if (lastfail != -4) log_print(lf, "ftdi_usb_open_dev: %s\n", ftdi_get_error_string(ftDevice));
        ftdi_free(ftDevice);
        ftDevice = NULL;
        return -4;
    }
    if (ftdi_usb_reset(ftDevice)) {
        ftdi_list_free(&list);
        if (lastfail != -5) log_print(lf, "ftdi_usb_reset: %s\n", ftdi_get_error_string(ftDevice));
        ftdi_free(ftDevice);
        ftDevice = NULL;
        return -5;
    }
    log_print(lf, "Served device: %s (%s)\n", serial, desc);
    ftdi_list_free(&list);

    ftdi_usb_purge_buffers(ftDevice);
    ftdi_set_latency_timer(ftDevice, 4);
#else
    DWORD n, i, j;
    FT_DEVICE_LIST_INFO_NODE *list;

    i_dev = dev;
    inited = 0;

    ftStatus = FT_CreateDeviceInfoList(&n);
    if (ftStatus != FT_OK) {
        if (lastfail != -6) log_print(lf, "FT_CreateDeviceInfoList failed\n");
        return -6;
    }
    if (n == 0) {
        if (lastfail != -3) log_print(lf, "No usable USB device found for now\n");
        return -3;
    }

    list = (FT_DEVICE_LIST_INFO_NODE*)malloc(n * sizeof(FT_DEVICE_LIST_INFO_NODE));
    if (!list) {
        if (lastfail != -1) log_print(lf, "Out of memory\n");
        return -1;
    }
    ftStatus = FT_GetDeviceInfoList(list, &n);
    if (ftStatus != FT_OK) n = 0;
    for (i = 0; i < n; i++) {
        if (dev) {
            if (!strcmp(dev, list[i].SerialNumber)) break;
        } else {
            if (!strcmp("IDE64 USB DEVICE", list[i].Description)) break;
        }
    }
    if (i == n && !dev) {
        for (i = 0; i < n; i++) {
            if (!strcmp("FT245R USB FIFO", list[i].Description)) break;
        }
    }
    if (i == n) {
        free(list);
        if (lastfail != -3) log_print(lf, "No usable USB device found for now\n");
        return -3;
    }
    ftStatus = FT_OpenEx((void*)list[i].LocId, FT_OPEN_BY_LOCATION, &ftHandle);
    if (ftStatus != FT_OK) ftStatus = FT_OpenEx(list[i].SerialNumber, FT_OPEN_BY_SERIAL_NUMBER, &ftHandle);
    if (ftStatus != FT_OK) {
        free(list);
        if (lastfail != -7) log_print(lf, "FT_OpenEx failed\n");
        return -7;
    }
    log_print(lf, "Served device: %s (%s)\n", list[i].SerialNumber, list[i].Description);
    for (j = 0; j < n; j++) {
        if (i == j) {
            continue;
        }
        log_print(lf, "Ignored device: %s (%s)\n", list[j].SerialNumber, list[j].Description);
    }
    free(list);

    FT_ResetDevice(ftHandle);
    FT_Purge(ftHandle, FT_PURGE_RX|FT_PURGE_TX);
    FT_SetTimeouts(ftHandle, 100, 10000);	// C64->PC, PC->C64
    FT_SetLatencyTimer(ftHandle, 1);
//     FT_SetUSBParameters(ftHandle, 64,0);
//     FT_SetDeadmanTimeout(ftHandle, 10);
#endif
    signal(SIGALRM, timeout);
    inited = 1;
    return 0;
}

static int getb(void) {
    unsigned char data;
    int err;

    alarm(1);
#ifndef WIN32
    do {
        err=ftdi_read_data(ftDevice, &data, 1);
        if (err == -ENODEV) return -ENODEV;
        if (err < 0) return -EIO;
        if (err == 1) {
            crc_add_byte(data);
            return (unsigned char)data;
        }
    } while (!tout);
#else
    DWORD dwBytesRead;

    do {
        err=FT_Read(ftHandle, &data, 1, &dwBytesRead);
        if (err == FT_DEVICE_NOT_FOUND) return -ENODEV;
        if (err != FT_OK) return -ENODEV;
        if (dwBytesRead == 1) {
            crc_add_byte(data);
            return (unsigned char)data;
        }
    } while (!tout);
#endif
    return -EIO;
}

static int getbytes(unsigned short int bytes, unsigned char *data) {
    int err = 1;
#ifdef WIN32
    DWORD dwBytesRead;
#endif
    do {
        int l;
        if (err) alarm(1);
        if (bytes>4096) l=4096; else l=bytes;
        if (tout) return -EIO;
#ifndef WIN32
        err=ftdi_read_data(ftDevice, data, l);
        if (err == -ENODEV) return -ENODEV;
        if (err < 0 ) return -EIO;
        crc_add_block(data, err);
        bytes-=err;data+=err;
#else
        err=FT_Read(ftHandle, data, l, &dwBytesRead);
        if (err == FT_DEVICE_NOT_FOUND) return -ENODEV;
        if (err != FT_OK) return -EIO;
        crc_add_block(data, dwBytesRead);
        bytes-=dwBytesRead;data+=dwBytesRead;
        err=(int)dwBytesRead;
#endif
    } while (bytes > 0);
    return 0;
}

static int sendb(unsigned char data) {
    if (ebufop >= sizeof(ebufo)) {
        if (flush()) return -EIO;
    }
    crc_add_byte(data);
    ebufo[ebufop++] = data;
    return 0;
}

static int sendbytes(unsigned short int bytes, unsigned char *data) {
    crc_add_block(data, bytes);
    while (ebufop + bytes > sizeof(ebufo)) {
        memcpy(ebufo + ebufop, data, sizeof(ebufo) - ebufop);
        data += sizeof(ebufo) - ebufop;
        bytes -= sizeof(ebufo) - ebufop;
        ebufop = sizeof(ebufo);
        if (flush()) return -EIO;
    }
    memcpy(ebufo + ebufop, data, bytes);
    ebufop += bytes;
    return 0;
}	

static void eshutdown(void) {
#ifndef WIN32
    if (ftDevice) {
        ftdi_usb_purge_buffers(ftDevice);
        ftdi_usb_close(ftDevice);
        ftdi_free(ftDevice);
        ftDevice = NULL;
    }
#else
    FT_Purge(ftHandle, FT_PURGE_RX | FT_PURGE_TX);
    FT_Close(ftHandle);
#endif
    inited = 0;
}

static int flush(void) {
    if (ebufop) {
        int err, l = ebufop;
        unsigned char *data = ebufo;
        ebufop = 0;
        do {
#ifdef WIN32
            DWORD dwBytesWritten;
            err = FT_Write(ftHandle, data, l, &dwBytesWritten);
            if (err == FT_DEVICE_NOT_FOUND) return -ENODEV;
            if (err != FT_OK) return -EIO;
            l -= dwBytesWritten; data += dwBytesWritten;
#else
            err = ftdi_write_data(ftDevice, data, l);
            if (err == -ENODEV) return -ENODEV;
            if (err < 0) return -EIO;
            l -= err; data += err;
#endif
        } while (l > 0);
    }
    return 0;
}

static void done(void) {
    alarm(0);
}

static int clean(void) {
#ifndef WIN32
    ftdi_usb_purge_buffers(ftDevice);
#else
    FT_Purge(ftHandle, FT_PURGE_RX | FT_PURGE_TX);
#endif
    return tout;
}

static int waitb(void) {
    int dyntime=1, err;
    unsigned char data;
    alarm(0);tout = 0;
    if (!inited) return -ENODEV;
    for (;;) {
#ifndef WIN32
        err=ftdi_read_data(ftDevice, &data, 1);
        if (err == -ENODEV) return -ENODEV;
        if (err < 0) return -EIO;
        if (err == 1) {
            crc_add_byte(data);
            return (unsigned char)data;
        }
#else
        DWORD dwBytesRead;

        err=FT_Read(ftHandle, &data, 1, &dwBytesRead);
        if (err == FT_DEVICE_NOT_FOUND) return -ENODEV;
        if (err != FT_OK) return -ENODEV;
        if (dwBytesRead == 1) {
            crc_add_byte(data);
            return (unsigned char)data;
        }
#endif
        usleep(dyntime);
        if (dyntime < 16384) dyntime<<=1;
    }
    return (unsigned char)data;
}

static int reinitialize(Log *lf, int lastfail) {
    eshutdown();
    return initialize(i_dev, 0, lf, lastfail);
}

const struct io_ops_s usb_io_ops = {
    .drivername   = drivername,
    .initialize   = initialize,
    .reinitialize = reinitialize,
    .getb         = getb,
    .sendb        = sendb,
    .getbytes     = getbytes,
    .sendbytes    = sendbytes,
    .shutdown     = eshutdown,
    .flush        = flush,
    .done         = done,
    .wait         = waitb,
    .clean        = clean,
};
